#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import math
import numpy as np

import collections
import heapq

import matplotlib.pyplot as plt

import Environment as env

LOCAL_NEIGHBORHOOD_SIZE = .1
COLLISION_RESOLUTION = .01

class SamplingPlanner:

    def __init__(self, limits):
        """
        Parameters
        ----------
        limits: list((float, float))
            translation limits in individual axes 
        """
        self.limits = limits
        self.max_norm_translation = np.max([self.limits[0][1]-self.limits[0][0],self.limits[1][1]-self.limits[1][0],self.limits[2][1]-self.limits[2][0], .0001]) * LOCAL_NEIGHBORHOOD_SIZE
        self.max_norm_rotation = np.max([self.limits[3][1]-self.limits[3][0],self.limits[4][1]-self.limits[4][0],self.limits[5][1]-self.limits[5][0], .0001]) * LOCAL_NEIGHBORHOOD_SIZE 

        self.collision_resolution_translation = np.max([self.limits[0][1]-self.limits[0][0],self.limits[1][1]-self.limits[1][0],self.limits[2][1]-self.limits[2][0], .0001]) * COLLISION_RESOLUTION
        self.collision_resolution_rotation = np.max([self.limits[3][1]-self.limits[3][0],self.limits[4][1]-self.limits[4][0],self.limits[5][1]-self.limits[5][0], .0001]) * COLLISION_RESOLUTION 

        print ('max norm: translation %.6f' % self.max_norm_translation)
        print ('max norm: rotation    %.6f' % self.max_norm_rotation)

        print ('collision resolution: translation %.6f' % self.collision_resolution_translation)
        print ('collision resolution: rotation    %.6f' % self.collision_resolution_rotation)

        # storage for sample plotting
        self.samples_vertices = np.zeros((3,0))
        self.samples_edges = []
        self.samples_path = []

    def is_valid_edge(self, start, end, environment, verbose):     
        """
        Is and oriented edge between two configurations valid? 

        Parameters
        ----------
        start: numpy array (4x4)
            start configuration of the robot in SE(3) coordinates
        end: numpy array (4x4)
            end configuration of the robot in SE(3) coordinates

        Returns
        -------
        boolean
            True if the edge is valid
        """
        assert False, 'This is an abstract method.'

    def plan(self, environment, start, goal):
        """
        Method to plan the path

        Parameters
        ----------
        environment: Environment
            Map of the environment that provides collision checking 
        start: numpy array (4x4)
            start configuration of the robot in SE(3) coordinates
        goal: numpy array (4x4)
            goal configuration of the robot in SE(3) coordinates

        Returns
        -------
        list(numpy array (4x4))
            the path between the start and the goal Pose in SE(3) coordinates
        """
        assert False, 'This is an abstract method.'

    def plot_samples(self, plot_samples = True, plot_edges = True, plot_path = True):
        """
        method to plot samples and solution into current figure
        """
        if plot_samples:
            xyz = self.samples_vertices
            plt.gca().scatter(xyz[0,:],xyz[1,:],xyz[2,:],c='b',s=2)
            plt.gca().scatter(xyz[0,:2],xyz[1,:2],xyz[2,:2],c='r',s=4)

        if plot_edges:    
            for (ti,tj) in self.samples_edges:
                c = 'b'
                plt.gca().plot([ti[0],tj[0]],[ti[1],tj[1]],[ti[2],tj[2]],c=c,linewidth=0.5)

        if plot_path:        
            for i in range(len(self.samples_path)-1):
                ti = self.samples_path[i]
                tj = self.samples_path[i+1]
                plt.gca().plot([ti[0],tj[0]],[ti[1],tj[1]],[ti[2],tj[2]],c='r',linewidth=2)


######################################################################################
# HELPER FUNCTIONS
######################################################################################

def e2r(e):
    """
    method to convert euler angles into rotation matrix
    @input: e .. euler [waw,pitch,roll]
    @output: r .. rotation matrix
    """
    Rx = [[1,0,0],
          [0,math.cos(e[0]),-math.sin(e[0])],
          [0,math.sin(e[0]),math.cos(e[0])]]
    Ry = [[math.cos(e[1]),0,math.sin(e[1])],
          [0,1,0],
          [-math.sin(e[1]),0,math.cos(e[1])]]
    Rz = [[math.cos(e[2]),-math.sin(e[2]),0],
          [math.sin(e[2]),math.cos(e[2]),0],
          [0,0,1]]
    r = np.matmul(Rz,np.matmul(Ry,Rx))
    return r

def r2e(r):
    """
    method to convert rotation matrix into euler angles
    @input: r .. rotation matrix
    @output: e .. euler [waw,pitch,roll]
    """
    x = math.atan2(r[2,1],r[2,2])
    y = math.atan2(-r[2,0],math.sqrt(r[2,1]*r[2,1]+r[2,2]*r[2,2]))
    z = math.atan2(r[1,0],r[0,0])
    e = [x, y, z]
    return e

def te_2_se3(te):
    """
    method to convert translation and euler angles into se3 matrix
    @input: te .. translation + euler angles [x,y,z,yaw,pitch,roll]
    @output: se3 .. matrix 4x4
    """
    t = te[:3]
    e = te[3:6]
    R = e2r(e)
    T = np.eye(4)
    T[:3,:3] = np.array(R)
    T[:3,3] = np.array(t)
    return T

def se3_2_te(se3):
    """
    method to convert se3 matrix into translation and euler angles
    @input: se3 .. matrix 4x4
    @output: te .. translation + euler angles [x,y,z,yaw,pitch,roll]
    """
    R = se3[:3,:3]
    e = r2e(R)
    t = se3[:3,3]
    te = [transl for transl in t] + e
    return te

def q2r(q):
    """
    method to convert quaternion into rotation matrix
    @input: q .. quaternion (qx,qy,qz,qw)
    @output: r .. rotation matrix
    """
    q = q[[3, 0, 1, 2]]
    r = [[q[0] ** 2 + q[1] ** 2 - q[2] ** 2 - q[3] ** 2, 2 * (q[1] * q[2] - q[0] * q[3]),
          2 * (q[1] * q[3] + q[0] * q[2])],
         [2 * (q[1] * q[2] + q[0] * q[3]), q[0] ** 2 - q[1] ** 2 + q[2] ** 2 - q[3] ** 2,
          2 * (q[2] * q[3] - q[0] * q[1])],
         [2 * (q[1] * q[3] - q[0] * q[2]), 2 * (q[2] * q[3] + q[0] * q[1]),
          q[0] ** 2 - q[1] ** 2 - q[2] ** 2 + q[3] ** 2]]
    return r

def r2q(r):
    """
    method to convert rotation matrix into quaternion
    @input: r .. rotation matrix
    @output: q .. quaternion (qx,qy,qz,qw)
    """
    q = np.array([r.trace() + 1, r[2, 1] - r[1, 2], r[0, 2] - r[2, 0], r[1, 0] - r[0, 1]])
    q = 1.0 / (2.0 * math.sqrt(r.trace() + 1.0000001)) * q
    q = q[[1, 2, 3, 0]]
    return q

def norm_translation(t1, t2):
    """
    method to compute 2nd norm for translation
    @input: t1,t2 .. poses [x,y,z]
    @output: float .. norm
    """
    return np.linalg.norm(np.array(t2) - np.array(t1))

def norm_rotation_R(R1,R2):
    """
    method to norm for rotation
    @input: R1,R2 .. rotation matrices 3x3
    @output: float .. norm
    """
    R = np.matmul(R1,R2.transpose())
    ang = np.arccos ( (np.trace(R) - 1.) / 2 )
    return ang

def slerp_se3(start, end, step_no):
    """
    method to compute spherical linear interpolation between se3 poses
    @input: start,end .. 4x4 se3 poses
    @input: step_no .. number of steps
    @output: steps .. list of the interpolated se3 poses, always at least [start,end]
    """
    t1 = start[:3,3]
    t2 = end[:3,3]
    R1 = start[:3,:3]
    R2 = end[:3,:3]

    q1 = r2q(R1)
    q2 = r2q(R2)

    thetaR1R2 = norm_rotation_R(R1,R2)/2

    steps = []

    if step_no < 2:
        steps = [start,end]
    else:
        for a in np.linspace(0.,1.,step_no+1):
            ta = t1 + (t2-t1)*a
            if np.abs(thetaR1R2) > .0001:
                qa = q1*np.sin( (1-a)*thetaR1R2 ) + q2*np.sin( a*thetaR1R2 ) / np.sin(thetaR1R2)
                qa = qa * (1/np.linalg.norm(qa))
                Ra = np.array(q2r(qa))
            else:
                Ra = R1
            Ta = np.eye(4)
            Ta[:3,:3] = Ra
            Ta[:3,3] = ta
            steps.append(Ta)

    return steps   

    
    

